SHELL=/bin/bash
################################################################################
#  This file is user-editable and allows configuration of the RNAstructure 
#  build system.
#
#     Contents:
#        Section 1: High-level configuration options.
#        Section 2: Compiler Settings, flags, and Recipes
#        Section 3: "Internal Stuff" - Utility functions and definitions.
#        Section 4: Operating System-Specific Configuration
#        Section 5: Common Linking and Compiling Rules
#        Section 6: Program-specific Linking and Compiling Recipes
################################################################################


################################################################################
#          Section 1: High-level configuration options
################################################################################
# The following variables can be set to change the behavior of this makefile.

# Environment variables: 
#   These are unset by default. But they can be set here or defined in the environment as desired.
#
#   RNA_MAKE_OS   -  If set, RNA_MAKE_OS will override the default
#                    operating system auto-detection.  Values: Mac, Windows or Linux
#   RNA_MAKE_ARCH -  If set, RNA_MAKE_ARCH will override the target machine 
#                    architecture (e.g. x86 or x86_64), e.g. for cross-compiling.

# Command-line variables: 
#   These are shorter forms for the above variables that can be passed on the 
#   command-line to make in the form "name=value"  e.g.:
#            $  make [TARGETS...]  debug=on arch=x86 
#
#   static=on|off        - force static or dynamic linking (default is OS-dependent)
#   dynamic=on|off       - opposite of static option above.
#   debug=on             - add debug flags (i.e. -g) to CXXFLAGS
#   optimize=off         - turn off optimization flags (i.e. -O3). Default is "on".
#   warnings=on          - turn on additional compiler warnings (e.g. -Wall)
#   os=Mac|Linux|Windows - compile for the specified operating system
#   arch=x86|x86_64      - compile for the specified machine architecture
#   quiet=on             - turn off a lot off make output and redirect stdout and stderr
#                          of compilation to files (named <TARGET>_make.log and <TARGET>_make-errors.log)
#                          This sets the QUIET variable that is used in COMPILE_CPP and LINK
#
################################################################################
##          Section 2: Compiler Settings and Flags
################################################################################
# These variables are common among several operating systems and compilers, but
# they can also be modified in the OS-specific setup in "Section 4".

# CXX           - Define the c++ compiler. 
#                 Examples: g++ (GNU C++ compiler), icc (Intel C compiler),
#                 CC (SGI and Sun compiler), xlC (xlc compiler, e.g. on AIX)
# CXXFLAGS      - Compiler flags that are common to all platforms and all builds.
#                 (Additional platform-specific or program-specific flags are appended below)
#                  GCC Options can be found here:  https://gcc.gnu.org/onlinedocs/gcc-4.9.2/gcc/Invoking-GCC.html#Invoking-GCC
#                 Do not include optimization flags (-O3) or debugging symbols (-g) here. 
#                 See DEBUG_FLAGS and OPTIMIZE_FLAGS below.
# COMPILE_CPP   - The default recipe to compile a .cpp source file into an object file.
#                 (Do not include $< because COMPILE_CPP is the base for other recipes that include it.)
  CXX = g++
  CXXFLAGS = -D NDEBUG ${ARCHITECTURE}  # Additional OS-dependent flags are added in the OS-specific section below.
  COMPILE_CPP = ${QUIET}${CXX} ${CXXFLAGS} -c -o $@
 
# DEBUG_FLAGS - compiler flag(s) to add debugging symbols. By default these are not used unless
#   make is invoked with the flag debug=yes ('true', '1', and 'on' are also acceptable).
  DEBUG_FLAGS= -g

# OPTIMIZE_FLAGS - compiler flag(s) to enable optimization. These WILL be added to CXXFLAGS 
#  unless disabled with the "optimize=off" argument to `make`.
  OPTIMIZE_FLAGS= -O3

# LINK          - Recipe for linking object files into an executable.
# LINK_LIBS     - List of additional libraries to pass to the linker. These are listed AFTER all object files (because order IS important, at least to GCC)
# STATIC_FLAGS  - Flags for compiling with static libraries. (Will only be used if DO_STATIC_LINK is "on").
# DYNAMIC_FLAGS - Flags for compiling with dynamic/shared libraries. (Will only be used if DO_STATIC_LINK is "").
# DO_STATIC_LINK- Set to "on" to enable static linking (by including STATIC_FLAGS in CXXFLAGS) 
#                 or leave it blank to use dynamic linking.  calling `make [TARGETS] static=on` 
#                 will also enable static linking.
  LINK = ${QUIET}${CXX} ${CXXFLAGS} $(if ${DO_STATIC_LINK},${STATIC_FLAGS},${DYNAMIC_FLAGS}) -o $@
  LINK_LIBS =   -lstdc++ -lpthread -lgomp 
  LINK_DEPS = ${LINK} $^ ${LINK_LIBS} # Shortcut for linking all listed dependencies with the standard recipe and libraries
  STATIC_FLAGS=-static -static-libgcc -static-libstdc++ -Wl,-Bstatic # seems redundant, but different versions of GCC require different flags
  DYNAMIC_FLAGS=  # can be modified in OS-dependent section
  DO_STATIC_LINK=

# LIBFLAGS   - compiler flags when producing a shared library (.dll, .so, etc)
  LIBFLAGS = -shared 

# CUDA_COMPILER - NVIDIA CUDA compiler, for GPU calculations
  CUDA_COMPILER = nvcc
# CUDAFLAGS - flags for the CUDA compiler. These will be merged with CXXFLAGS using -Xcompiler.
  CUDAFLAGS = -D FLOAT -D SHORT -D _FORCE_INLINES -O3 -use_fast_math  -Wno-deprecated-gpu-targets -D _CUDA_CALC_ -D NDEBUG
  CUDA_DEBUG_FLAGS=-g

################################################################################
##           Section 3: Utility functions and OS-detection etc.
################################################################################
#  This section has no configuration options. It just defines some 
#  utility/convenience functions that are used later on, and it performs  
#  operating-system detection and other automation stuff. 
#  There are additional utility functions and variable definitions in Section 7, but
#  *this* section must appear here because it is used by subsequent sections (e.g. Section 4)

# The config file can be created to contain additional, directory or user-specific 
# variable definitions. The `make autoconfig` command can create this file with 
# default or auto-detected settings.
CONFIG_FILE=compiler.makecfg
-include $(CONFIG_FILE)

# IS_CMD_SET - function that evaluates the variable name passed in as the first
# argument, but ONLY if the specified flag was set by the user on the make command-line.
# Usage:  $(call IS_CMD_SET,<FLAGNAME>)
  IS_CMD_SET=$(if $(findstring command line, $(origin $(1))),$($(1)))
  
# IS_CMD_SET_ON is a function that evaluates to non-empty only if the specified 
# flag was passed on the make command-line and it evaluates to "1", "yes", "true", or "on".
# Usage:  $(call IS_CMD_SET_ON,<FLAGNAME>)
  IS_CMD_SET_ON=$(if $(call IS_CMD_SET,$(1)),$(filter yes on true 1,$($(1))))
  IS_CMD_SET_OFF=$(if $(call IS_CMD_SET,$(1)),$(filter off no false 0,$($(1))))

#### Detection of command-line flags passed to make: ####
# detect whether the "debug" flag is set to "on"
  ifneq ($(call IS_CMD_SET_ON,debug),)
    $(info Building with debugging symbols.)
    #removed override
    CXXFLAGS+=$(DEBUG_FLAGS)
    CUDAFLAGS+=$(CUDA_DEBUG_FLAGS)
    #export CXXFLAGS
  endif

# detect whether the "static" or "dynamic" flags were set on the command-line to make.
  ifneq ($(call IS_CMD_SET,static)$(call IS_CMD_SET,dynamic),)
    # TODO: if the user sets dynamic or static to some invalid value, the results may be unexpected. Add some code here to catch this and show an error to the user. e.g. (if dynamic isn't trueish or falseish, show an error. do the same for static)
    # set the DO_STATIC_LINK variable to "on" if `static` is "true-ish" (yes, on, true or 1) OR dynamic is "false-ish" (no off false or 0)
    # override means that even if the OS-dependent section tries to modify DO_STATIC_LINK, it will retain the value set here.
    override DO_STATIC_LINK:=$(if $(call IS_CMD_SET_ON,static)$(call IS_CMD_SET_OFF,dynamic),on)
    export DO_STATIC_LINK
    $(info Using $(if $(DO_STATIC_LINK),static,dynamic) linking.)
  endif

# detect whether the "optimize" flag is set to "off"
  ifneq ($(call IS_CMD_SET_OFF,optimize),)
    $(info Building without optimization.)
  else
    CXXFLAGS+=$(OPTIMIZE_FLAGS)
  endif

# detect whether the "warnings" flag is set to "off"
  ifneq ($(call IS_CMD_SET_ON,warnings),)
    $(info Building with additional warnings.)
    DISABLE_WARNINGS=
    CXXFLAGS+=-Wall
  endif

# detect whether the "quiet" flag is set to "on"
  ifneq ($(call IS_CMD_SET_ON,quiet),)
    fnMkFileName=$(subst .,_,$(subst /,_,$1))
    QUIET=@echo "Compiling $@"; >"$(call fnMkFileName,$@)_make.log" 2>"$(call fnMkFileName,$@)_make-errors.log" #    
  else
    QUIET=
  endif

# detect whether the "os" flag was set from the command-line
  ifneq ($(call IS_CMD_SET,os),)
    override RNA_MAKE_OS=$(os)
    export RNA_MAKE_OS
  endif

# detect whether the "arch" flag was set from the command-line
  ifneq ($(call IS_CMD_SET,arch),)
    override RNA_MAKE_ARCH=$(arch)
    export RNA_MAKE_ARCH
  endif


# Operating-System and Processor Architecture Detection
# This section attempts to auto-detect the OS. This is used only for convenience
# and allows the same Makefile to be used on multiple operating systems
# without modification.
# You can circumvent auto-detection by setting these environment variables:
#   RNA_MAKE_OS   -- the operating system name (e.g. Linux, Windows, Mac)
#   RNA_MAKE_ARCH -- the target architecture (e.g. x86 or x86_64)
# (These can be set as environment variables or on the MAKE command-line. 
#    e.g. `make all RNA_MAKE_OS=Linux`)
    ifneq (${RNA_MAKE_OS},) 
      #if RNA_MAKE_OS is NOT blank, use it as the OS
      OPSYSTEM=$(RNA_MAKE_OS)
    else ifeq (${OPSYSTEM},)
      # IF both RNA_MAKE_OS and OPSYSTEM are blank, use the "OSTYPE"
      # environment variable for detection. 
      OSTYPE := $(shell echo $$OSTYPE)
      ifneq (${OSTYPE},)
        # Perform some replacements to normalize the value of OSTYPE on various systems.
        OSTYPE := $(OSTYPE:cygwin%=Windows)
        OSTYPE := $(OSTYPE:msys=Windows)
        OSTYPE := $(OSTYPE:darwin%=Mac)# e.g. darwin14
        OSTYPE := $(OSTYPE:linux%=Linux)# e.g. linux-gnu
        OPSYSTEM=$(OSTYPE)
      else 
        $(warning The OSTYPE environment variable is missing. Please inform the developers.)
        # PREVIOUS method was to use Use the `uname` command to determine the OS.
        # This might still be required for some operating systems
        #   OPSYSTEM:=$(shell uname -s 2>/dev/null || echo UNKNOWN)#   Replace 'UNKNOWN' with default OS if desired. 
        # Perform some replacements to normalize the output of uname on various systems.
        #OPSYSTEM := $(OSTYPE:cygwin=Windows)
        #OPSYSTEM := $(OSTYPE:CYGWIN%=Windows)
        #OPSYSTEM := $(OPSYSTEM:MSYS%=Windows)
        #OPSYSTEM := $(OPSYSTEM:Darwin=Mac)
        #OPSYSTEM := $(OPSYSTEM:GNU%=Linux)
      endif
      $(if $(DEBUG), $(info Make: Operating System: $(OPSYSTEM)))
      export OPSYSTEM #make it available for recursive calls to make, so auto-detection is performed only once.
    endif
#
# The target architecture (i.e. x86 or x86_64) is probably that of the current OS,
# however the RNA_MAKE_ARCH variable can be used to cross-compile for a different 
# architecture if supported by the compiler etc.
    ifneq (${RNA_MAKE_ARCH},) #processor arch. 
      TARGET_ARCH=$(RNA_MAKE_ARCH)
    else ifeq (${TARGET_ARCH},) 
      # Determine processor architecture of the the OS.
      # If desired, you may replace the term 'UNKNOWN' with the default 
      # architecture (i.e. x86 or x86_64)
      TARGET_ARCH := $(shell echo $$HOSTTYPE)
      #PREVIOUS: TARGET_ARCH := shell uname -m 2>/dev/null || echo UNKNOWN)   
      $(if $(DEBUG), $(info Make: Architecture: $(TARGET_ARCH)))
      export TARGET_ARCH #make it available for recursive calls to make, so auto-detection is performed only onces
    endif
# Determine the number of bits (32 or 64) of the target architecture. This is used to name some files (see library_defines.h for details.)
  ARCH_BITS = $(if $(findstring 64,$(TARGET_ARCH)),64,32)
  NATIVE_ARCH_BITS := $(if $(findstring 64,$(shell echo $$HOSTTYPE)),64,32)
  
# Verify that the ROOTPATH variable is defined and is correct. Otherwise set it automatically.
# (ROOTPATH is the relative or absolute path to the RNAstructure root directory)
ifeq (${ROOTPATH},) # could also verify using $(wildcard ${ROOTPATH}/src/version.h) or $(wildcard ${ROOTPATH}/$(basename $(lastword $(MAKEFILE_LIST)))))
  # Get the parent directory of the last file included by the top-most Makefile
  # Then remove final slash (for aesthetics)
  ROOTPATH:=$(patsubst %/,%,$(dir $(lastword $(MAKEFILE_LIST)))) # MAKEFILE_LIST contains the names of processed Makefiles, with *THIS* one last e.g. "Makefile compiler.h" or "Makefile ../compiler.h" etc
endif

################################################################################
#           Section 4: Operating-System Specific Configuration
################################################################################
    ifeq (${OPSYSTEM},Linux)
      ############ LINUX ##########################################################
      MULTIFINDFLAG = -DMULTIFIND #### this flag does not seem to be used anywhere
      CXXFLAGS+= -fPIC
      DYNAMIC_FLAGS=-Wl,--as-needed # "as-needed" forces linker to ignore listed dlls that arent really necessary
      # Add -m32 if compiling for 32-bit on a 64-bit machine.
      ifneq ($(NATIVE_ARCH_BITS),$(ARCH_BITS))
        ARCHITECTURE = -m$(ARCH_BITS)
      endif
    else ifeq (${OPSYSTEM},Mac)
      ############ MAC ############################################################
      DISABLE_WARNINGS=-Wno-write-strings -Wno-unused-value -Wno-c++11-compat-deprecated-writable-strings -Wno-logical-op-parentheses -Wno-parentheses-equality
      CXXFLAGS+= -fPIC -arch $(TARGET_ARCH) $(DISABLE_WARNINGS)
      LIBFLAGS = -dynamiclib -arch $(TARGET_ARCH)
      DO_STATIC_LINK=## Do not use static linking on mac. Apple strongly discourages it.
      DYNAMIC_FLAGS= # The --as-needed flag is not supported by clang
      LINK_LIBS=-lstdc++ # Do not include -lphreads or -lgomp on Mac
    else ifeq (${OPSYSTEM},Windows)
      ############ WINDOWS ########################################################
      #  The following settings are for compilation on Windows using the MinGW-w64 
      #  compiler from within the Cygwin POSIX environment. Other choices of shell
      #  or compiler may work as well. 
      #  Compiler executable names may be dependent on the flavor of MinGW 
      #  you have installed. The names below correspond to the "MinGW-w64" build 
      #  obtained directly from the Cygwin repository. Note that 32-bit and 64-bit 
      #  targeted versions of the compiler can be installed side by side.
      WIN64-g++ = x86_64-w64-mingw32-g++
      WIN32-g++ = i686-w64-mingw32-g++
      CXX=$(if $(findstring 64,${TARGET_ARCH}),$(WIN64-g++),$(WIN32-g++))
      CXXFLAGS+= -fsched-spec-load -D__USE_MINGW_ANSI_STDIO=1 -pthread ## -fPIC is not supported on windows
      #On Windows, statically link libgcc, libstdc++, winpthreads, libgomp (OpenMP). 
      DO_STATIC_LINK=on# This can be overriden by the `static=` flag to make.
      # Read about static and shared linking on Windows: https://access.redhat.com/documentation/en-US/Red_Hat_Enterprise_Linux/4/html/Using_ld_the_GNU_Linker/win32.html
      #STATIC_FLAGS=-static -static-libgcc -static-libstdc++ -Wl,-Bstatic -lstdc++ -lpthread -lgomp 
      # INCLUDE_GAMMA : When compiling on windows with certain 
      #   compilers (Intel, VS etc), we need to include a file to redefine the 
      #   'tgamma' function that is missing from cmath. 
      #   The variable INCLUDE_GAMMA is used to indicate this.
      INCLUDE_GAMMA = $(if $(findstring g++,$(CXX)),,${ROOTPATH}/src/gamma.o)

      DYNAMIC_FLAGS=-Wl,--as-needed # "as-needed" forces linker to ignore listed dlls that arent really necessary

      # DYNALIGN_BIG_STACK
      # Compiling with MinGW-w64 on Windows seems to cause dynalign to crash (stack overflow)
      # unless the stack size is increased (with a linking option)
      DYNALIGN_BIG_STACK= -Wl,--stack,8388608
    else ifeq (,${OPSYSTEM})
      ############ NO OS ########################################################
      $(error No Operating system defined!!!)
    else
    ############ UNKNOWN OS ###################################################
      $(error Unknown Operating system defined: $(OPSYSTEM))
    endif

################################################################################
#       Section 5: Common Linking and Compiling Rules
################################################################################
# Default Compilation Rules - generic rules for compiling c++ object files (.o) from
# source files (cpp). These rules will be overridden if a more specific rule 
# is defined (see dependencies.h) 
# Add a rule for cpp files with corresponding header files:
%.o: %.cpp %.h
	$(MAKE_OUTDIR)
	${COMPILE_CPP} $<

# If no header file is found, this second rule will apply:
%.o: %.cpp
	$(MAKE_OUTDIR)
	${COMPILE_CPP} $<

## TODO: make the above rules more strict by prepending $(ROOTPATH) (but this means that other Makefiles have to specify the full sub-path to each .o instead of a relative path from the current directory. e.g. see PARTS/Makefile )


################################################################################
#       Section 6: Program-Specific Linking and Compiling Flags
################################################################################
# This section contains compilation and linking recipes and definitions that
# are used in specific circumstances or with specific programs.
# Usually nothing below this point needs to be changed by end-users, unless certain 
# standard compiler flags need to be modified for a given compiler (e.g. -c or -o )
# It may also be useful to look at the commands below for reference.

  # Define compiling and linking rules for OpenMP programs:
  COMPILE_OMP=${COMPILE_CPP} ${CXXOPENMPFLAGS}
  CXXOPENMPFLAGS = -fopenmp -D SMP # g++ compiler
  #CXXOPENMPFLAGS = -openmp -D SMP # intel compiler
  LINK_SMP = ${LINK} ${CXXOPENMPFLAGS}
  LINK_SMP_DEPS = ${LINK_SMP} $^ ${LINK_LIBS}

  # Define the SMP compiling rule (for SMP without OpenMP -- e.g. Dynalign, Turbofold).
  COMPILE_SMP = ${COMPILE_CPP}  -D COMPILE_SMP

  # Define a Dynalign II compiling rule.
  COMPILE_DYNALIGN_II = ${COMPILE_CPP} -D DYNALIGN_II

  # Define a Dynalign II SMP compiling rule.
  COMPILE_DYNALIGN_II_SMP = ${COMPILE_CPP} -D DYNALIGN_II -D COMPILE_SMP

  # Define an MULTIFIND compiling rule.
  COMPILE_MULTIFIND = ${COMPILE_CPP} -D MULTIFIND

  # Define CUDA compiling and linking rules (for use with GPU programs).
  COMPILE_CUDA = nvcc ${CUDAFLAGS} -c -o $@
  LINK_CUDA = nvcc ${CUDAFLAGS} -o $@

  # Define an SVM compiling rule.
  COMPILE_SVM = ${COMPILE_CPP} -I${INCLUDESVM}
  # Define an SVM/SMP compiling rule.
  COMPILE_SVM_SMP = ${COMPILE_CPP} -I${INCLUDESVM} -D COMPILE_SMP
  # If making Multifind, INCLUDESVM must indicate the location of svm.o from the libsvm package. 
  INCLUDESVM = /usr/local/libsvm-3.17

  # Define an INSTRUMENTED compiling rule.
  COMPILE_INSTRUMENTED = ${COMPILE_CPP} -D INSTRUMENTED


######################### GPU/CUDA ############################
# The CUDA environment variable indicates that RNAstructure programs should
# use the CUDA versions of partition and fold.
# This requires building the "normal" classes with the CUDA compiler (nvcc) and
# CUDA flags as well as compiling the CUDA-specific versions of partition and Fold.
ifneq (${CUDA},)
  $(info CUDA environment variable detected: compiling for GPU)
  CXX=${CUDA_COMPILER} # use nvcc to compile "normal" RNAstructure code.
  DYNAMIC_FLAGS=# --as-needed flag is not supported by nvcc
  CXXFLAGS := $(CUDAFLAGS) $(patsubst %,-Xcompiler %,$(CXXFLAGS)) # pass the CUDA flags as well as the standard CXX compiler flags.
endif

##############################################################################
# Section 7:  Debugging and Utility Macros/Variables
##############################################################################
# Some useful definitions for special characters
#newline (linefeed) character
define newline


endef

comma:= ,
empty:=
space:= $(empty) $(empty)
tab:= $(empty)	$(empty)
dollar:=$$
hash:=\#
bkslash:=\\
colon:=:
t:=$(tab)
n:=$(newline)
_:=$(empty)
s:=$(space)
k:=$(bkslash)
h:=$(hash)

# MKDIR: A function to make a directory.
#      Usage: $(call MKDIR,PATH_TO_DIR)
MKDIR = @[[ -e $(1) ]] || mkdir -p $(1) &> /dev/null

# MAKE_OUTDIR: A macro to use in recipes. 
#    It simply makes the parent directory of the target file.
#    Usage: $(MAKE_OUTDIR)
MAKE_OUTDIR = $(call MKDIR,$(@D))

############# Debugging ######################################################
# Show information about the operating system and configured variables if
# debug output has been enabled.
#############################################################################
fShowVar=$(1)$(2)Source:$(3)$(origin $(1))$(2)Define:$(3)$(value $(1))$(2)Value :$(3)$($(1))
#fShowVarList=$(foreach var,$(1),HZ@)#$(call fShowVar,$(var),M,X)
#fShowVarList=$(foreach var,$(1),$(call fShowVar,$(var),$(newline)      ,$(tab)))
fShowVarList=$(foreach var,$(1),$n  $(call fShowVar,$(var),$n    ,$s))
fSpacesToLines=$(subst $s,$n,$(1))
#fShowVarListOld=$(foreach var,$(1),\#    $(call fShowVar,$(var),$(newline)\#       ,$(tab))$(newline))\#

CPP_MAKE_VARS=COMPILE_CPP CXX CXXDEFINES CXXFLAGS LIBFLAGS LINK DO_STATIC_LINK STATIC_FLAGS OPSYSTEM TARGET_ARCH 
APP_MAKE_VARS=COMPILE_CUDA COMPILE_DYNALIGN_II COMPILE_DYNALIGN_II_SMP COMPILE_INSTRUMENTED  COMPILE_MULTIFIND COMPILE_SMP COMPILE_SVM COMPILE_SVM_SMP CUDA_COMPILER CUDAFLAGS CXXOPENMPFLAGS INCLUDE_GAMMA INCLUDESVM LINK_CUDA LINK_SMP MULTIFINDFLAG 
JAVA_MAKE_VARS=COMPILE_JNI JNI_CXXFLAGS JNI_DIR JNI_INCLUDE_FLAGS JNI_INCLUDE JXX JXXFLAGS LINK_JNI_LIB

showvars:
	$(info \
$n---------------------------- Makefile Configuration ----------------------------\
$n  Building on OS: $(OPSYSTEM) (uname: $(OPSYSTEM_RAW))\
$n  Target Architecture: $(TARGET_ARCH) ($(ARCH_BITS)-bits)\
$n$n-----------C++ Configuration------------$(call fShowVarList,$(CPP_MAKE_VARS))\
$n$n-----------Java Configuration-----------$(call fShowVarList,$(JAVA_MAKE_VARS))\
$n$n-----------App-specific Configuration-----------$(call fShowVarList,$(APP_MAKE_VARS))\
$n$n       (For a list of ALL variables, use the target 'showallvars')\
$n--------------------------------------------------------------------------------)
showallvars:
	$s  $(foreach var,$(sort $(.VARIABLES)),$(info   $(call fShowVar,$(var),$n   ,$s)$n))

############# Optional auto-Configuration ##################################################
# Define the config file here so make doesn't complain if it is missing.
$(CONFIG_FILE):

export CP_UPD?=cp # cp command

# autoconfig - Run some tests and save some settings.
autoconfig:
	@echo 'Makefile Auto-Config'
	@echo '## This is an auto-generated file. It may not be wise to edit it.' > $(CONFIG_FILE)
	@echo 'RNA_MAKE_OS=$(RNA_MAKE_OS)' >> $(CONFIG_FILE)
	@echo 'RNA_MAKE_ARCH=$(RNA_MAKE_ARCH)' >> $(CONFIG_FILE)
	@# Test for cp -u 
	@$(shell touch '#1.tmp')
	@CP_UPD=cp$(shell cp -u '#1.tmp' '#2.tmp' &> /dev/null && echo ' -u')
	@echo 'CP_UPD=$($CP_UPD)' >> $(CONFIG_FILE)
	@$(shell rm -f '#?.tmp') 
	@echo 'Auto-Config Complete. Values written to $(abspath $(CONFIG_FILE)).'

#Clear the default goal if it was set to one of the targets in this file. (It should be set in the Makefile itself.)
ifeq ($(.DEFAULT_GOAL),showvars)
   .DEFAULT_GOAL=
endif

##############################################################################
# Disable built-in rules. All the rules we need are specified explicitly.
MAKEFLAGS += --no-builtin-rules --no-builtin-variables
.SUFFIXES:
####### Define default message/action for files with no matching rule. #######
.DEFAULT:
	$(error Unknown target "$@".  For a list of goals, type "make help".)
##############################################################################
